/* RemoteLoggingServices.cs
 * The public interface of the remote call logging framework. Also implements the tracking
 * handler.
 *
 * Distributed with no license what-so-ever - feel free to do
 * whatever the hell you want with this :-)  I would appreciate it
 * if you dropped me a line to let me know if you've found this
 * useful though.
 *
 * Have fun!
 *
 * Tomer Gabel
 * http://www.tomergabel.com
 * tomer@tomergabel.com
 *
 */

using System;
using System.Collections;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Services;

namespace TomerGabel.RemoteLogging
{
	public sealed class RemoteLoggingServices
	{
		/// <summary>
		/// Provides a .NET Remoting track handler for the remote logging infrastructure
		/// </summary>
		private class TrackingHandler : ITrackingHandler
		{
			/// <summary>
			/// Notifies the current instance that an object has been marshaled.
			/// </summary>
			/// <param name="obj"></param>
			/// <param name="or">The <see cref="T:System.Runtime.Remoting.ObjRef"/> that results from marshaling and represents the specified object.</param>
			public void MarshaledObject( object obj, ObjRef or )
			{
				RemoteLoggingServices.m_trackedUris[ or.URI ] = obj;
			}

			/// <summary>
			/// Notifies the current instance that an object has been unmarshaled.
			/// </summary>
			/// <param name="obj"></param>
			/// <param name="or">The <see cref="T:System.Runtime.Remoting.ObjRef"/> that represents the specified object.</param>
			public void UnmarshaledObject( object obj, ObjRef or )
			{
				RemoteLoggingServices.m_trackedUris.Remove( or.URI );
			}

			/// <summary>
			/// Notifies the current instance that an object has been disconnected from its proxy.
			/// </summary>
			/// <param name="obj">The disconnected object.</param>
			public void DisconnectedObject( object obj )
			{
			}
		}
		
		
		private static Hashtable m_trackedUris = Hashtable.Synchronized( new Hashtable() );
		private static Hashtable m_consumers   = new Hashtable();
		
		/// <summary>
		/// Initializes the <see cref="RemoteLoggingServices"/> class.
		/// </summary>
		static RemoteLoggingServices()
		{
			TrackingServices.RegisterTrackingHandler( new TrackingHandler() );
		}
		
		/// <summary>
		/// Registers the specified consumer to track all remote method calls.
		/// </summary>
		/// <param name="consumer">The consumer.</param>
		public static void Register( IRemoteLoggingConsumer consumer )
		{
			lock ( m_consumers.SyncRoot )
				m_consumers[ consumer ] = null;
		}
		/// <summary>
		/// Registers the specified consumer to track method calls for a specific type.
		/// </summary>
		/// <param name="consumer">The consumer.</param>
		/// <param name="track">The type to track.</param>
		public static void Register( IRemoteLoggingConsumer consumer, Type track )
		{
			lock ( m_consumers.SyncRoot )
				m_consumers[ consumer ] = track;
		}
		/// <summary>
		/// Unregisters the specified consumer.
		/// </summary>
		/// <param name="consumer">The consumer.</param>
		public static void Unregister( IRemoteLoggingConsumer consumer )
		{
			lock ( m_consumers.SyncRoot )
				m_consumers.Remove( consumer );
		}

		/// <summary>
		/// Notifies the remote logging infrastructure that a particular method has been called on
		/// a remote object
		/// </summary>
		/// <param name="uri">Uri for the remote object</param>
		/// <param name="serverType">Implementation server type</param>
		/// <param name="method">The method called on the remoted object</param>
		/// <param name="args">Method call arguments</param>
		public static void Notify( string uri, Type serverType, string method, object[] args )
		{
			// Normalize URI...
			if ( !uri.StartsWith( "/" + RemotingConfiguration.ApplicationId ) )
				uri = "/" + RemotingConfiguration.ApplicationId + uri;
			
			object target = m_trackedUris[ uri ];
			
			if ( null != target )
			{
				lock ( m_consumers.SyncRoot )
					foreach ( DictionaryEntry consumer in m_consumers )
					{
						if ( ( null == consumer.Value ) || ( serverType == consumer.Value ) )
							( (IRemoteLoggingConsumer) consumer.Key ).HandleRemoteCall( target, method, args );
					}
				target.GetType();
			}
		}
	
		/// <summary>
		/// Begins method call tracking
		/// </summary>
		public static void Track()
		{
			// Doesn't really do anything - just makes sure the static ctor is called...
		}
	}
}
