/* RemoteLoggingSink.cs
 * Provides the remoting server channel sink for the remote logging framework.
 *
 * Distributed with no license what-so-ever - feel free to do
 * whatever the hell you want with this :-)  I would appreciate it
 * if you dropped me a line to let me know if you've found this
 * useful though.
 *
 * Have fun!
 *
 * Tomer Gabel
 * http://www.tomergabel.com
 * tomer@tomergabel.com
 *
 */

using System;
using System.Collections;
using System.Collections.Specialized;
using System.IO;
using System.Reflection;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Channels;
using System.Runtime.Remoting.Messaging;

namespace TomerGabel.RemoteLogging
{
	/// <summary>
	/// Provides the implementation for the Remoting Logging Sink
	/// </summary>
	public class RemoteLoggingSink : BaseChannelSinkWithProperties, IServerChannelSink
	{
		#region Ctor
		/// <summary>
		/// Initializes a new instance of the <see cref="RemoteLoggingSink"/> class.
		/// </summary>
		/// <param name="next">The next server channel sink in the sink chain.</param>
		public RemoteLoggingSink( IServerChannelSink next )
		{
			m_next = next;
		}

		#endregion
		
		#region IServerChannelSink implementation
		private IServerChannelSink m_next;

		/// <summary>
		/// Requests message processing from the current sink.
		/// </summary>
		/// <param name="sinkStack">A stack of channel sinks that called the current sink.</param>
		/// <param name="requestMsg">The message that contains the request.</param>
		/// <param name="requestHeaders">Headers retrieved from the incoming message from the client.</param>
		/// <param name="requestStream">The stream that needs to be to processed and passed on to the deserialization sink.</param>
		/// <param name="responseMsg">When this method returns, contains an <see cref="T:System.Runtime.Remoting.Messaging.IMessage"/> that holds the response message. This parameter is passed uninitialized.</param>
		/// <param name="responseHeaders">When this method returns, contains an <see cref="T:System.Runtime.Remoting.Channels.ITransportHeaders"/> that holds the headers that are to be added to return message heading to the client. This parameter is passed uninitialized.</param>
		/// <param name="responseStream">When this method returns, contains a <see cref="T:System.IO.Stream"/> that is heading back to the transport sink. This parameter is passed uninitialized.</param>
		/// <returns>
		/// A <see cref="T:System.Runtime.Remoting.Channels.ServerProcessing"/> status value
		/// that provides information about how message was processed.
		/// </returns>
		/// <exception cref="T:System.Security.SecurityException">The immediate caller does not have infrastructure permission.</exception>
		public ServerProcessing ProcessMessage( IServerChannelSinkStack sinkStack, IMessage requestMsg,
		                                        ITransportHeaders requestHeaders, Stream requestStream,
		                                        out IMessage responseMsg, out ITransportHeaders responseHeaders,
		                                        out Stream responseStream )
		{
			sinkStack.Push( this, null );

			ServerProcessing res = m_next.ProcessMessage(
				sinkStack, requestMsg, requestHeaders, requestStream, 
				out responseMsg, out responseHeaders, out responseStream 
				);
			
			if ( null == requestMsg )
				throw new Exception( "Request message is null; try changing the order of the " +
									 "sink providers so that the RemoteLoggingSinkProvider " +
									 "comes after the formatter in the chain." );

			if ( requestMsg is IMethodCallMessage )
				processIncomingMethodcall( (MethodCall) requestMsg );
			
			return res;
		}

		/// <summary>
		/// Requests processing from the current sink of the response from a method call
		/// sent asynchronously.
		/// </summary>
		/// <param name="sinkStack">A stack of sinks leading back to the server transport sink.</param>
		/// <param name="state">Information generated on the request side that is associated with this sink.</param>
		/// <param name="msg">The response message.</param>
		/// <param name="headers">The headers to add to the return message heading to the client.</param>
		/// <param name="stream">The stream heading back to the transport sink.</param>
		/// <exception cref="T:System.Security.SecurityException">The immediate caller does not have infrastructure permission.</exception>
		public void AsyncProcessResponse( IServerResponseChannelSinkStack sinkStack, object state, IMessage msg,
		                                  ITransportHeaders headers, Stream stream )
		{
			sinkStack.AsyncProcessResponse( msg, headers, stream );
		}

		/// <summary>
		/// Returns the <see cref="T:System.IO.Stream"/> onto which the provided
		/// response message is to be serialized.
		/// </summary>
		/// <param name="sinkStack">A stack of sinks leading back to the server transport sink.</param>
		/// <param name="state"></param>
		/// <param name="msg"></param>
		/// <param name="headers"></param>
		/// <returns>
		/// The <see cref="T:System.IO.Stream"/> onto which the provided
		/// response message is to be serialized.
		/// </returns>
		/// <exception cref="T:System.Security.SecurityException">The immediate caller does not have infrastructure permission.</exception>
		public Stream GetResponseStream( IServerResponseChannelSinkStack sinkStack, object state, IMessage msg,
		                                 ITransportHeaders headers )
		{
			return null;
		}

		/// <summary>
		/// Gets the next server channel sink in the server sink chain.
		/// </summary>
		/// <value></value>
		public IServerChannelSink NextChannelSink
		{
			get { return m_next; }
		}

		#endregion
		
		#region Caching
		/// <summary>
		/// Provides a cache for methods implemented by a particular server type
		/// </summary>
		private class MethodCache : DictionaryBase
		{
			private Type m_serverType;
			
			/// <summary>
			/// Initializes a new instance of the <see cref="MethodCache"/> class.
			/// </summary>
			/// <param name="serverType">The server type for this cache</param>
			public MethodCache( Type serverType )
			{
				m_serverType = serverType;
			}
			
			/// <summary>
			/// Checks and returns if the specified method is flagged for logging.
			/// </summary>
			/// <value></value>
			public bool this[ MethodBase method ]
			{
				get
				{
					return (bool) this.InnerHashtable[ method ];
				}
			}
			
			
			/// <summary>
			/// Determines whether the cache contains the specified method.
			/// </summary>
			/// <param name="method">The method to look up.</param>
			/// <returns>
			/// 	<c>true</c> if the cache contains the specified method; otherwise, <c>false</c>.
			/// </returns>
			public bool Contains( MethodBase method )
			{
				return this.InnerHashtable.ContainsKey( method );
			}
			
			
			/// <summary>
			/// Adds the specified method to the cache.
			/// </summary>
			/// <param name="method">The method to be added to the cache.</param>
			/// <returns><c>true</c> if the method is flagged for logging, <c>false</c>
			/// otherwise</returns>
			public bool Add( MethodBase method )
			{
				MethodInfo target = null;
				if ( method.DeclaringType.IsInterface )
				{
					// Cool trick by Andy Smith (http://weblogs.asp.net/asmith/archive/2003/07/11/10005.aspx)
					InterfaceMapping map = m_serverType.GetInterfaceMap( method.DeclaringType );
					for ( int i = 0; i < map.InterfaceMethods.Length; ++i )
						if ( method.Name == map.InterfaceMethods[ i ].Name )
						{
							target = map.TargetMethods[ i ];
							break;
						}
				}
				else
				{
					// Ugly hack to look up implemented method call
					ParameterInfo[] paraList = method.GetParameters();
					Type[] typeList = new Type[ paraList.Length ];
					for ( int i = 0; i < paraList.Length; ++i )
						typeList[ paraList[ i ].Position ] = paraList[ i ].ParameterType;
					
					target = m_serverType.GetMethod( 
						method.Name, 
						BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic,
						null,
						typeList,
						null 
						);
				}

				bool flag = ( target == null ) ? false : target.IsDefined( typeof( LogRemoteCallAttribute ), true );
				this.InnerHashtable[ method ] = flag;
				return flag;
			}
		}
	
		/// <summary>
		/// Provides a cache for server types and their associated methods
		/// </summary>
		private class TypeCache : DictionaryBase
		{
			/// <summary>
			/// Gets the <see cref="MethodCache"/> for the specified type.
			/// </summary>
			/// <value>The <see cref="MethodCache"/> for the specified type.</value>
			public MethodCache this[ Type type ]
			{
				get
				{
					return (MethodCache) this.InnerHashtable[ type ];
				}
			}
			
			/// <summary>
			/// Adds the specified type to the type cache.
			/// </summary>
			/// <param name="type">The type to be added.</param>
			/// <returns>A MethodCache instance for this type</returns>
			public MethodCache Add( Type type )
			{
				MethodCache cache = new MethodCache( type );
				this.InnerHashtable.Add( type, cache );
				return cache;
			}
			
			/// <summary>
			/// Determines whether the type cache contains the specified type.
			/// </summary>
			/// <param name="type">The type to be looked up</param>
			/// <returns>
			/// 	<c>true</c> if the type cache ccontains the specified type; otherwise, <c>false</c>.
			/// </returns>
			public bool Contains( Type type )
			{
				return this.InnerHashtable.ContainsKey( type );
			}
		}
		/// <summary>
		/// Provides a cache for information pertaining to server URIs recognized by this provider.
		/// </summary>
		private class UriCache : NameObjectCollectionBase
		{
			/// <summary>
			/// A single entry in the URI cache
			/// </summary>
			public class UriEntry
			{
				private bool m_flagged;
				private Type m_serverType;

				/// <summary>
				/// Initializes a new instance of the <see cref="UriEntry"/> class.
				/// </summary>
				/// <param name="serverType">The implementation server type for this URI</param>
				public UriEntry( Type serverType )
				{
					if ( null == serverType )
						throw new ArgumentNullException( "serverType" );
					
					m_serverType = serverType;
					m_flagged = serverType.GetCustomAttributes( typeof( LogRemoteCallAttribute ), true ).Length > 0;
				}

				
				/// <summary>
				/// Gets a value indicating whether the implementation server type for this URI
				/// is flagged for logging.
				/// </summary>
				/// <value><c>true</c> if the implementation server type for this URI is flagged
				/// for logging; otherwise, <c>false</c>.</value>
				public bool Flagged
				{
					get { return m_flagged; }
				}

				/// <summary>
				/// Gets the implementation server type for this URI
				/// </summary>
				/// <value>The implementation server type for this URI</value>
				public Type ServerType
				{
					get { return m_serverType; }
				}
			}

			/// <summary>
			/// Gets the <see cref="UriEntry"/> with the specified URI.
			/// </summary>
			/// <value></value>
			public UriEntry this[ string uri ]
			{
				get
				{
					return (UriEntry) this.BaseGet( uri );
				}
			}
			
			/// <summary>
			/// Determines whether the URI cache contains the specified URI.
			/// </summary>
			/// <param name="uri">The URI to look up</param>
			/// <returns>
			/// 	<c>true</c> if the URI cache contains the specified URI; otherwise, <c>false</c>.
			/// </returns>
			public bool Contains( string uri )
			{
				return this.BaseGet( uri ) != null;
			}
			
			/// <summary>
			/// Adds the specified URI to the URI cache.
			/// </summary>
			/// <param name="uri">The URI to be added</param>
			/// <returns>A new <see cref="UriEntry"/> instance for the specified URI</returns>
			public UriEntry Add( string uri )
			{
				UriEntry entry = new UriEntry( RemotingServices.GetServerTypeForUri( uri ) );
				this.BaseAdd( uri, entry );
				return entry;
			}
		}
			
		private UriCache m_uriCache = new UriCache();
		private TypeCache m_typeCache = new TypeCache();
		#endregion
		
		#region Message processing
		/// <summary>
		/// Processes the incoming method call
		/// </summary>
		/// <param name="call">Method call to be processed</param>
		private void processIncomingMethodcall( IMethodCallMessage call )
		{
			if ( null == call )
				throw new ArgumentNullException( "call" );

			// Look for call URI in cache. If not found, get server type for call URI and add to
			// cache
			UriCache.UriEntry uriEntry = 
				m_uriCache.Contains( call.Uri ) ? 
				m_uriCache[ call.Uri ] :
				m_uriCache.Add( call.Uri );

			bool flag = uriEntry.Flagged;
			if ( !uriEntry.Flagged )
			{
				MethodCache methods =
					m_typeCache.Contains( uriEntry.ServerType ) ?
					m_typeCache[ uriEntry.ServerType ] :
					m_typeCache.Add( uriEntry.ServerType );
					
				// Check if the server implementation for the given method call is flagged for
				// logging. If necessary, find out and add to cache
				flag |=
					methods.Contains( call.MethodBase ) ?
					methods[ call.MethodBase ] :
					methods.Add( call.MethodBase );
			}
			
			if ( flag )
				RemoteLoggingServices.Notify( call.Uri, uriEntry.ServerType, call.MethodName, call.Args );
		}
		#endregion
	}
}
